#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

extern int *solver_main(char *);

int sol[9*9];

char *buffer;
int size;
int maxsize;

void add_char(char c)
{
  if (size == maxsize) {
    buffer = realloc(buffer, maxsize += 2048);
    if (!buffer) { fprintf(stderr, "out of memory\n"); exit(1); }
  }
  buffer[size] = c;
  size++;
}

void add_int(int c)
{
  int i;
  int len;
  char s[64];

  sprintf(s, "%d", c);
  len = strlen(s);

  for (i=0; i<len; i++)
    add_char(s[i]);
}

void add_clause(int *c, int n)
{
  int i;
  for (i=0; i<n; i++) {
    add_int(c[i]);
    add_char(' ');
  }
  add_int(0);
  add_char('\n');
}

/*
 * add clauses describing val(i) != v
 */
void add_negative(int i, int v)
{
  int c[4];
  i *= 4;
  i++;
  /*
   * X !=i means:
   *    X0 != i0 OR X1 != i1 ... OR X3 != i3
   * Xn != in translates to Xn if in==0 and to !Xn if in==1
   * So we have a simple clause here.
   * In the code below:
   * i means Xn
   * -i means !Xn
   * We have done i*=4 because variable i has boolean variables i*4 ... i*4+3
   * and i++ because DIMACS starts to count from 1 not 0
   */
  c[0] = (v&1)? -i : i; i++;
  c[1] = (v&2)? -i : i; i++;
  c[2] = (v&4)? -i : i; i++;
  c[3] = (v&8)? -i : i;
  add_clause(c, 4);
}

/*
 * add clauses describing val(i) == v
 */
void add_positive(int i, int v)
{
  int c;
  i *= 4;
  i++;
  /* X = i means:
   * X0 = i0 AND X1 = i1 ... AND X3 = i3
   * Xn = in translates to Xn if in==1 and to !Xn if in==0
   * So we have 4 unity clauses here.
   * In the code below:
   * i means Xn
   * -i means !Xn
   * We have done i*=4 because variable i has boolean variables i*4 ... i*4+3
   * and i++ because DIMACS starts to count from 1 not 0
   */
  c = (v&1)? i : -i; i++; add_clause(&c, 1);
  c = (v&2)? i : -i; i++; add_clause(&c, 1);
  c = (v&4)? i : -i; i++; add_clause(&c, 1);
  c = (v&8)? i : -i; add_clause(&c, 1);
}

/*
 * add clauses describing val(i) != val(j)
 */
void add_var_diff(int i, int j)
{
  int c[8];
  int l;
  i *= 4;
  j *= 4;

  i++;
  j++;

  /* we have 2 vars: X and Y
   * that have 4 bits, so we have X1, X2, X3, X4 and Y1, Y2, Y3, Y4
   * X != Y means:
   *    X1 != Y1 OR X2 != Y2 OR X3 != Y3 OR X4 != Y4
   * X1 != Y1 means (X1 AND !Y1) OR (!X1 AND Y1) is true
   * We want a CNF form for that, so we apply the wonderful rule:
   *    (A AND B) OR C = (A OR C) AND (B OR C)
   * which gives an exponential blow up, but we don't care here.
   * We finally get by multiple applications of the rule that:
   *   X1 != Y1 means (X1 OR Y1) AND (!X1 OR !Y1)
   * We have the same for X2 != Y2 etc.
   * Now we have:
   *   ((X1 OR Y1) AND (!X1 OR !Y1)) OR ((X2 OR Y2) AND (!X2 OR !Y2)) OR
   *   ((X3 OR Y3) AND (!X3 OR !Y3)) OR ((X4 OR Y4) AND (!X4 OR !Y4))
   * We still have to apply the same rule to take out the AND and take
   * in the OR.
   * We finally find 16 clauses, of shape:
   *   T1i(X1, Y1) OR T2i(X2, Y2) OR T4i(X3, Y3) OR T8i(X4, Y4)
   * where Tji is true if bit j of integer i is 1
   * and when Tji is true then Tji(Xn, Yn) = Xn OR Yn
   * and when Tji is false then Tji(Xn, Yn) = !Xn OR !Yn
   * This loop does create the 16 clauses by using this Tji stuff.
   * variable ii to ii+3 represent the 4 bits of X
   * variable jj to jj+3 represent the 4 bits of Y
   * ii means Xn
   * -ii means !Xn
   * We have done i*=4 because variable i has boolean variables i*4 ... i*4+3
   * and i++ because DIMACS starts to count from 1 not 0
   * (in this code we first put the Xn boolean variables followed
   * by the Yn ones)
   */
  for (l=0; l<16; l++) {
    int ii = i;
    int jj = j;
    int *cc = c;

    *cc++ = (l&1) ? ii : -ii; ii++;
    *cc++ = (l&2) ? ii : -ii; ii++;
    *cc++ = (l&4) ? ii : -ii; ii++;
    *cc++ = (l&8) ? ii : -ii;

    *cc++ = (l&1) ? jj : -jj; jj++;
    *cc++ = (l&2) ? jj : -jj; jj++;
    *cc++ = (l&4) ? jj : -jj; jj++;
    *cc++ = (l&8) ? jj : -jj;

    add_clause(c, 8);
  }
}

void force_value(int i)
{
  /* we handle 9x9 sudoku, so a variable can take values 1..9
   * which means it cannot take values 0 and 10..15
   * (we code on 4 bits, so we may have values 0..15)
   * which translates as:
   * X != 0 AND X != 10 AND ... AND X != 15
   * Each inequality has then to be put as a CNF clause.
   */
  add_negative(i, 15);
  add_negative(i, 14);
  add_negative(i, 13);
  add_negative(i, 12);
  add_negative(i, 11);
  add_negative(i, 10);
  add_negative(i, 0);
}

/* return 0 if no solution, 1 if solution */
int sat_solve()
{
  int i;
  int *ret;

  /* we put a trailing 0 in the buffer */
  add_char(0);
  ret = solver_main(buffer);
  /* and remove the trailing 0 because we will add clauses for next run */
  size--;

  if (!ret) {
    printf("no solution\n");
    return 0;
  }

  for (i=0; i<9*9; i++) {
    int res = 0;
    /* to get the value of variable i is simply to put the 4 corresponding
     * bits together
     * The code below does exactly that.
     * For i, its boolean variables are i*4 ... i*4 + 3
     */
    if(ret[i*4+0] == 1) res |= 1;
    if(ret[i*4+1] == 1) res |= 2;
    if(ret[i*4+2] == 1) res |= 4;
    if(ret[i*4+3] == 1) res |= 8;
    printf("%d%c", res, i % 9 == 8 ? '\n' : ' ');
    sol[i] = res;
  }

  free(ret);
  return 1;
}

int cube_to_var(int cube, int v)
{
  /* variables in a cube are numbered as follow:
   * 0 1 2
   * 3 4 5
   * 6 7 8
   * little cubes in the big cube is also numbered this way.
   * This leads to the following code. We first get the column
   * and line of the little cube, then deduce its topleft
   * variable, and then we find the x and y in the little cube
   * of v and we finally get the variable we seek based on that.
   */
  int x_cube = cube % 3;
  int y_cube = cube / 3;
  int top_left = y_cube * 9 * 3 + x_cube * 3;
  int x_v = v % 3;
  int y_v = v / 3;
  return top_left + y_v * 9 + x_v;
}

/*
 * this function forces variables i and j in the cube 'cube'
 * to be different
 */
void do_cube(int cube, int i, int j)
{
  int v1 = cube_to_var(cube, i);
  int v2 = cube_to_var(cube, j);
  add_var_diff(v1, v2);
}

void forbid_sol(void)
{
  int c[9*9*4];
  int *cc = c;
  int i;

  /* to forbid a solution means that
   * X0 != sol0 OR X1 != sol1 OR ... OR X81 != sol81
   * then each Xi != soli is transformed as in the function
   * add_negative.
   * So we finally have a big clause of 9*9*4 literals.
   * The code below does that.
   * it will add ii != v in each loop.
   * ii will be the 4 bits of variables i and v the solution of cell i.
   * (the code is clear, no?)
   */
  for (i=0; i<9*9; i++) {
    int v = sol[i];
    int ii = i*4+1;
    *cc++ = (v&1)? -ii : ii; ii++;
    *cc++ = (v&2)? -ii : ii; ii++;
    *cc++ = (v&4)? -ii : ii; ii++;
    *cc++ = (v&8)? -ii : ii;
  }

  add_clause(c, 9*9*4);
}

void add_negative_values(int var_1, int value_1, int var_2, int value_2)
{
  int c[8];
  int i;

  /* add the clause (var_1 != value_1) or (var_2 != value_2)
   * we have similar stuff than add_negative
   */

  i = var_1 * 4 + 1;
  c[0] = (value_1&1)? -i : i; i++;
  c[1] = (value_1&2)? -i : i; i++;
  c[2] = (value_1&4)? -i : i; i++;
  c[3] = (value_1&8)? -i : i;

  i = var_2 * 4 + 1;
  c[4] = (value_2&1)? -i : i; i++;
  c[5] = (value_2&2)? -i : i; i++;
  c[6] = (value_2&4)? -i : i; i++;
  c[7] = (value_2&8)? -i : i;

  add_clause(c, 8);
}

void add_diff(int dif, int left, int right)
{
  int i, j;
  /* so we want left = right + dif or right = left + dif
   * it means that we don't want left != right + dif
   * or abs(left-right) != dif
   * if we pick up all possible values for left and right,
   * this results in the following statement:
   * not(left=X and right=Y) and not(left=Z and right=ZZ) and ...
   * each not(left=X and right=Y) becomes:
   * not (left0 and left1 and left2 and left3 and right0 and ... and right3)
   * with each leftX and rightX of the form vi or !vi (a variable or
   * its negation)
   * if you put the not inside, you've got a clause.
   * thus the code below.
   *
   * Note that directly encoding left = right + dif would
   * mean stuff like (left=X and right=Y) or (left=Z and right=ZZ) or ...
   * which needs to be translated into CNF, which explodes like hell!
   *   We obtain millions of clauses because each (left=X and right=Y)
   * has 8 variables and we can have up to 16 such assignments
   * (the case where dif=1), which leads to 8^16 clauses, which is
   * quite huge, no? (we use the formula:
   * (A and B) or C = (A or C) and (B or C)).
   * (If you don't get the point, contact me, sed@free.fr)
   */
  for (i=1; i<=9; i++)
    for (j=1; j<=9; j++)
      if (abs(i-j) != dif)
        add_negative_values(left, i, right, j);
}

void add_sup(int left, int right)
{
  int i, j;
  /* we want left = right + 1
   * so we don't want left != right + 1
   * we derive the code below
   */
  for (i=1; i<=9; i++)
    for (j=1; j<=9; j++)
      if (i != j+1)
        add_negative_values(left, i, right, j);
}

void add_inf(int left, int right)
{
  int i, j;
  /* we want left = right - 1
   * so we don't want left != right - 1
   * we derive the code below
   */
  for (i=1; i<=9; i++)
    for (j=1; j<=9; j++)
      if (i != j-1)
        add_negative_values(left, i, right, j);
}

int main(int n, char **vv)
{
  int i;
  int j;
  int ret;
  int nb_sol;
  int line;
  int column;
  int cube;
  int cur_case;

  if (n != 2) {
    printf("usage: mdudidoku <dudidoku file> ('-' for stdin)\n");
    printf("  The file has the following structure:\n");
    printf("    for each line of the grid, you have 8 constraints\n");
    printf("    (one constraint for each couple of adjacent cells)\n");
    printf("    a constraint can be:\n");
    printf("      *   : means no constraint\n");
    printf("      >   : means the '>1' constraint\n");
    printf("      *   : means the '<1' constraint\n");
    printf("      + X : means values from left and right must differ of X\n");
    printf("  See dudidoku.0 for an example describing the problem at:\n");
    printf("    http://plus.maths.org/issue38/puzzle/index.html\n");
    printf("  You may use any space you want between numbers and symbols.\n");
    printf("http://sed.free.fr/complex/sudoku.html\n");
    return 0;
  }

  if (vv[1][0] != '-' || vv[1][1] != 0) {
    stdin = fopen(vv[1], "r");
    if (!stdin) {
      perror(vv[1]);
      return 0;
    }
  }

  /* let's force the cases to have values between 1 and 9 inclusively */
  for (i=0; i<9*9; i++)
    force_value(i);

  /* AND all vars on a line must be != */
  /* wonderful for for for */
  for (line=0; line<9; line++)
    for (i=0; i<9; i++)
      for (j=i+1; j<9; j++)
        add_var_diff(i+line*9, j+line*9);

  /* AND all vars on a column must be != */
  for (column=0; column<9; column++)
    for (i=0; i<9; i++)
      for (j=i+1; j<9; j++)
        add_var_diff(i*9+column, j*9+column);

  /* AND all vars in a cube must be != */
  for (cube=0; cube<9; cube++)
    for (i=0; i<9; i++)
      for (j=i+1; j<9; j++) {
        /* don't add already added stuff (same line or same column) */
        if (i/3 == j/3) continue;
        if (i%3 == j%3) continue;
        do_cube(cube, i, j);
      }

  /* AND read a duplex difference sudoku, which means to constrain some
   * values
   */
  cur_case = 0;
#define INC { cur_case++; if (cur_case % 9 == 8) cur_case++; }
  while (1) {
    int c = fgetc(stdin);
    if (c == -1) break;
    if (isspace(c)) continue;
    if (c == '*') {
      INC; continue;
    }
    if (c == '+') {
      int dif, case_left, case_right;
      if (fscanf(stdin, "%d", &dif) != 1) {
        fprintf(stderr, "bad duplex difference sudoku\n");
        return 1;
      }
      case_left = cur_case;
      case_right = cur_case + 1;
      if (dif<2 || dif>8 || case_left >= 9*9 || case_right >= 9*9) {
        fprintf(stderr, "bad duplex difference sudoku\n");
        return 1;
      }
      add_diff(dif, case_left, case_right);
      INC; continue;
    }
    if (c == '>') {
      int case_left, case_right;
      case_left = cur_case;
      case_right = cur_case + 1;
      if (case_left >= 9*9 || case_right >= 9*9) {
        fprintf(stderr, "bad duplex difference sudoku\n");
        return 1;
      }
      add_sup(case_left, case_right);
      INC; continue;
    }
    if (c == '<') {
      int case_left, case_right;
      case_left = cur_case;
      case_right = cur_case + 1;
      if (case_left >= 9*9 || case_right >= 9*9) {
        fprintf(stderr, "bad duplex difference sudoku\n");
        return 1;
      }
      add_inf(case_left, case_right);
      INC; continue;
    }
    fprintf(stderr, "bad duplex difference sudoku\n");
    return 1;
  }
  nb_sol = 0;

  /* solve it to death, until no more solution is found */
  /* unlike zchaff, minisat does not add
   * clauses on the fly, so the more solutions you get,
   * the slower the solver, because there are more and more
   * clauses parsed at each run.
   */
  do {
    printf("checking new solution\n");
    ret = sat_solve();
    if (ret) {
      /* if there was a solution, we forbid the solver to return it the
       * next time we run it
       */
      forbid_sol();
    }
    nb_sol += !!ret;
  } while (ret);

  printf("%d solutions\n", nb_sol);

  return 0;
}
